## deuNet Package
deuNet is a highly modular neural networks library, written in Python and capable of running on top of Theano. It is developed with a focus on enabling 
fast experimentation. Use deuNet if you need a deep learning library that:
    - allows easy and fast prototyping (through total modularity minimalism, and extensibitylity).
    - supports construction of convolutional networks, multi-layer perceptions
    - supports basic NN modules such as affine layer, conv layer, pooling module
    - supports modern NN components including maxout, dropout
    - supports various network architectures, e.g. MLP, Auto-encoder, Convnet, Recurrent net.
    - supports GPU usage
For better interpretation Ipython notebooks are also included demonstrating some state-of-art NN architectures, such as Alex Net, VGG net, LSTM, and etc.

## Installation
deuNet uses the following dependencies:
    - numpy, scipy
    - recommended if use CNNs on NVIDIA GPUs: cuDNN
    - Theano

To install Theano, see this page:
http://deeplearning.net/software/theano/install.html

For documentation and source code of Theano, see the project website:
http://deeplearning.net/software/theano/

## Usage
To use deuNet library, one would follow the workflow as below:
    - load a dataset, either through a provided function in deuNet/datasets or through a function of your own
    - build a NN model container
    - add layers
    - train the model
    - test the model and get performance

Example is as follows.
'''
    # load data
    (train_X, train_y), (test_X, test_y) = mnist.load_data() # load MNIST dataset.

    # Define NN structure
    model = NN()

    # convnet layers
    model.add(Convolution2D(32,1,3,3, border_mode='full',
                            init='glorot_uniform',activation='relu', reg_W=0))
    model.add(Convolution2D(32,32,3,3, border_mode='valid',
                            init='glorot_uniform',activation='relu', reg_W=0))
    model.add(MaxPooling2D(pool_size=(2,2)))

    # fully connected layers
    model.add(Flatten())
    model.add(AffineLayer(32*196,128,init='glorot_uniform',activation='relu',reg_W=0))
    model.add(AffineLayer(128,10,init='glorot_uniform',activation='softmax',reg_W=0))

    # Compile NN
    print 'Compile ConvNet ...'
    model.compile(optimizer='SGD', loss='categorical_crossentropy',
            reg_type='L2', learning_rate=learning_rate, momentum=momentum,
            lr_decay=lr_decay, nesterov=nesterov, rho=rho)

    # Train NN
    model.fit(train_X, train_y, valid_X, valid_y,
            batch_size=batch_size, nb_epoch=nb_epoch, verbose=True)

    # Test NN
    model.get_test_accuracy(test_X, test_y)
'''

## Directory Layout
deuNet (current directory) is the distribution directory.
    * deuNet/deuNet contains the package source code
    * deuNet/deuNet has several submodules:
	* activations.py implements a number of activation functions, see below for details 
	* initializations.py implements a number of common weight initialization methods, see below for details
	* losses.py implements common loss functions used for weight update, see below for details
	* models.py implements the NN class (only sequential NN is supported as of now), see below for details
	* optimizers.py implements several optimization methods, each as a class that contains update functions, see below for details
	* regularizers.py implements regularization functions for controlling weight decay.
	* layers/ contains a number of common layers, including convolution, batch norm, etc. see below for details
	* datasets/ contains data loading functions for public benchmark data including cifar10, cifar100, mnist, etc.
	* preprocessing/ contains steps of processing images and texts prior to entering neural networks
	* utils/ includes general utility functions
    * deuNet/demos has several demonstration scripts for using the deuNet package
	The file name of each script indicates (data)_(network). For example, cifar_10_alex.py demonstrates building an
	AlexNet on CIFAR 10 dataset. 
 
## Demos
A total of 14 demo scripts are included in the deuNet/demo directory. Each demo script is self-contained. For example, mnist_mlp.py 
is a script building MLP network on MNIST data.
You can run this demo by setting a few Theano flags before starting python, like this:
	THEANO_FLAGS=mode=FAST_RUN,device=gpu,floatX=float32 python mnist_mlp.py
Another example:
 	THEANO_FLAGS=mode=FAST_RUN,device=gpu,floatX=float32 python mnist_cnn.py

## Guiding principles

- Modularity: A model is understood as a sequence of layers, borrowed the concept from Torch. Allow all parts (layers, optimizers, ...) to be used independently of deuNet.

- Simplicity: Be easy to use, easy to understand and easy to extend in research. For example, a 2 layer MLP network can be written as simply as:
    ``` Python
    model = NN() 
    model.add(AffineLayer(D, nb_hidden1, init='glorot_uniform',activation='relu', reg_W = reg_W))
    model.add(AffineLayer(nb_hidden1, nb_classes, init='glorot_uniform',activation='softmax',reg_W=reg_W))
    ```

- Out-of-box optimization: supports multiple gradient based optimizers. One line of code can be used to plug the optimizer into defined model, for example:
    ``` Python
    model.compile(optimizer='RMSprop', loss='categorical_crossentropy', 
                  reg_type='L2', learning_rate = learning_rate, momentum=momentum,
                  lr_decay=lr_decay, nesterov=nesterov, rho=rho)v
    ```

- Easy training process: borrowed the API of sklearn, easy to plug in training and validation data for training. No need for iterative training loop.
    ``` Python
    model.fit(train_X, train_y, valid_X, valid_y,
              batch_size=batch_size, nb_epoch=nb_epoch, verbose=True)
    ```


## Supported Layers
The layers folder contains the definition of all kinds of Neural network layers (building blocks).

- AffineLayer: fully connected layer.

- Dropout: dropout to avoid overfitting. Reference: http://www.cs.toronto.edu/~rsalakhu/papers/srivastava14a.pdf

- Flatten: flatten a multiple dim tensor into 1-D tensor.

- Convolution2D: 2D convolution network, which use Nvidia CuDNN as backend.

- MaxPooling2D: pooling (sampling) for convolution layer.

- Convolution1D: 1D convolution network, which use Nvidia CuDNN as backend.

- MaxPooling1D: pooling (sampling) for convolution layer.

- BatchNormalization: batch norm implementation. Reference: http://arxiv.org/pdf/1502.03167v3.pdf


## Supported Neuron Activations
acitvations.py contains common used activations for non-linear transformations. It contains: 

- softmax
- relu
- tanh
- sigmoid
- hard_sigmoid
- linear
- leaky_relu
- maxout


## Supported Optimizers
The optimizers for NN are gradient computation and parameter updates methods. Once the analytic gradient is computed with backprogation (by theano), the gradients are used to perform a parameter update.

I implemented several approachs to perform the weights updating.

Reference: http://cs231n.github.io/neural-networks-3/#sgd

- SGD: vanillia version of gradient descent method.

- PMSprop: a very effective, but unpublished adaptive learning rate method. It is introduec by Hinton in his Coursera class. Reference: http://www.cs.toronto.edu/~tijmen/csc321/slides/lecture_slides_lec6.pdf

- Adagrad: It is an adaptive learning rate method proposed by J. Duchi.  It is very similar with RMSprop, but with no extra hyper-parameters. Reference: http://www.jmlr.org/papers/volume12/duchi11a/duchi11a.pdf

- Adadelta: another relatively common adaptive learning rate method.
    Reference: http://arxiv.org/abs/1212.5701


## Supported Loss functions
In the training process, once we compute the output of a neurual network by input a training set (or mini-batch), in order to adjust the parameter, performance needs to be evaluated. classification accuracy is one measurement. However, accuracy is non-differentiable. Some better measurements have been commonly used, such as cross-entropy for classification problem and L2-distance for regression problem.

- categorical_crossentropy: used for classification problems.

- euclidean_loss: Reference: http://mlg.eng.cam.ac.uk/yarin/blog_3d801aa532c1ce.html 

- mean_squared_error: used for regression.

# License
Copyright (c) 2015 Rosanne Liu, CUCIS, Northwestern University.
